"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""

# noinspection PyUnresolvedReferences
import ujson
import logging
from contextlib import closing

from gevent.threading import Lock
from pysolmeters.Meters import Meters

from pysolmysql.Pool.mysql_pool import MysqlConnectionPool

logger = logging.getLogger(__name__)


class MysqlApi(object):
    """
    Mysql Api
    """

    # Lock
    POOL_LOCK = Lock()

    # Static pool instances (hash from config dict => MysqlConnectionPool)
    D_POOL_INSTANCES = dict()

    @classmethod
    def reset_pools(cls):
        """
        Reset all pools
        """

        with cls.POOL_LOCK:
            for s_hash, pool in cls.D_POOL_INSTANCES.items():
                logger.info("Closing pool, s_hash=%s", s_hash)
                pool.close_all()
            cls.D_POOL_INSTANCES = dict()

    @classmethod
    def _get_pool_hash(cls, conf_dict):
        """
        Get pool hash
        :param conf_dict: dict
        :type conf_dict: dict
        :return: str
        :rtype: str
        """

        s_hash = str(hash(ujson.dumps(conf_dict, sort_keys=True)))
        return s_hash

    @classmethod
    def _get_pool(cls, conf_dict):
        """
        Init static pool
        :param conf_dict: dict
        :type conf_dict: dict
        :return pysolmysql.Pool.mysql_pool.MysqlConnectionPool
        :rtype pysolmysql.Pool.mysql_pool.MysqlConnectionPool
        """

        # Hash
        s_hash = cls._get_pool_hash(conf_dict)

        # Alloc if needed
        if s_hash not in cls.D_POOL_INSTANCES:
            with cls.POOL_LOCK:
                if s_hash not in cls.D_POOL_INSTANCES:
                    cls.D_POOL_INSTANCES[s_hash] = MysqlConnectionPool(conf_dict)
                    logger.info("Allocated pool, s_hash=%s, pool.len=%s", s_hash, len(cls.D_POOL_INSTANCES))
                    Meters.aii("k.db_pool.hash.cur")

        # Over
        return cls.D_POOL_INSTANCES[s_hash]

    @classmethod
    def _fix_type(cls, data):
        """
        Fix type
        :param data: data
        """
        if isinstance(data, bytearray):
            return data.decode("utf-8")
        else:
            return data

    @classmethod
    def exec_0(cls, conf_dict, statement):
        """
        Execute a sql statement, returning nothing.
        :param conf_dict: configuration dict
        :type conf_dict: dict
        :param statement: statement to execute
        :type statement: str
        """

        cnx = None
        try:
            cnx = cls._get_pool(conf_dict).connection_acquire()
            with closing(cnx.cursor()) as cur:
                cur.execute(statement)
        finally:
            cls._get_pool(conf_dict).connection_release(cnx)

    @classmethod
    def exec_n(cls, conf_dict, statement, fix_types=True):
        """
        Execute a sql statement, returning 0..N rows
        :param conf_dict: configuration dict
        :type conf_dict: dict
        :param statement: statement to execute
        :type statement: str
        :param fix_types: If true, fix data type
        :type fix_types: bool
        :return list of dict.
        :rtype list
        """

        cnx = None
        try:
            cnx = cls._get_pool(conf_dict).connection_acquire()
            with closing(cnx.cursor()) as cur:
                cur.execute(statement)
                rows = cur.fetchall()
                for row in rows:
                    logger.debug("row=%s", row)
                    for k, v in row.items():
                        logger.debug("k=%s, %s, %s", k, type(v), v)
                        if fix_types:
                            row[k] = MysqlApi._fix_type(v)
                return rows
        finally:
            cls._get_pool(conf_dict).connection_release(cnx)

    @classmethod
    def exec_1(cls, conf_dict, statement, fix_types=True):
        """
        Execute a sql statement, returning 1 row.
        Method will fail if 1 row is not returned.
        :rtype: object
        :param conf_dict: configuration dict
        :type conf_dict: dict
        :param statement: statement to execute
        :type statement: str
        :param fix_types: If true, fix data type
        :type fix_types: bool
        :return dict
        :rtype dict
        """

        cnx = None
        try:
            cnx = cls._get_pool(conf_dict).connection_acquire()
            with closing(cnx.cursor()) as cur:
                cur.execute(statement)
                rows = cur.fetchall()
                for row in rows:
                    logger.debug("row=%s", row)
                    for k, v in row.items():
                        logger.debug("k=%s, %s, %s", k, type(v), v)
                        if fix_types:
                            row[k] = MysqlApi._fix_type(v)
                if len(rows) != 1:
                    raise Exception("Invalid row len, expecting 1, having={0}".format(len(rows)))
                return rows[0]
        finally:
            cls._get_pool(conf_dict).connection_release(cnx)

    @classmethod
    def exec_01(cls, conf_dict, statement, fix_types=True):
        """
        Execute a sql statement, returning 0 or 1 row.
        Method will fail if 0 or 1 row is not returned.
        :param conf_dict: configuration dict
        :type conf_dict: dict
        :param statement: statement to execute
        :type statement: str
        :param fix_types: If true, fix data type
        :type fix_types: bool
        :return dict, None
        :rtype dict, None
        """

        cnx = None
        try:
            cnx = cls._get_pool(conf_dict).connection_acquire()
            with closing(cnx.cursor()) as cur:
                cur.execute(statement)
                rows = cur.fetchall()
                for row in rows:
                    logger.debug("row=%s", row)
                    for k, v in row.items():
                        logger.debug("k=%s, %s, %s", k, type(v), v)
                        if fix_types:
                            row[k] = MysqlApi._fix_type(v)
                if len(rows) == 0:
                    return None
                elif len(rows) != 1:
                    raise Exception("Invalid row len, expecting 1, having={0}".format(len(rows)))
                else:
                    return rows[0]
        finally:
            cls._get_pool(conf_dict).connection_release(cnx)

    @classmethod
    def multi_n(cls, conf_dict, ar_statement):
        """
        Execute multiple sql statement, reading nothing from mysql.
        :type conf_dict: dict
        :param ar_statement: list of statements to execute (for instance, batch of insert or whatever)
        :type ar_statement: list
        """

        cnx = None
        try:
            cnx = cls._get_pool(conf_dict).connection_acquire()
            with closing(cnx.cursor()) as cur:
                for s in ar_statement:
                    cur.execute(s)
        finally:
            cls._get_pool(conf_dict).connection_release(cnx)



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""
import copy
import logging
import os
import random
import struct

import pymysql
import time
from pysolbase.SolBase import SolBase
from pysolmeters.Meters import Meters

from pysolmysql.Pool.base_pool import DatabaseConnectionPool

logger = logging.getLogger(__name__)

# Init random
random.seed(a=struct.unpack('i', os.urandom(4)))


class MysqlConnectionPool(DatabaseConnectionPool):
    """
    Mysql connection pool, gevent compliant.

    This support multiple target hosts.

    In multiple target hosts mode, this is intended to be used with a mariadb active/active galera cluster.
    """

    # Public sample config dict
    # FOR : pysolmysql.Pool.mysql_pool.MysqlConnectionPool#create_connection
    # FOR : pysolmysql.Mysql.MysqlApi.MysqlApi
    PUBLIC_SAMPLE_CONFIG_DICT = {
        # HIGH PRIO
        "hosts": ["127.0.0.1", "127.0.0.1"],
        # LOW PRIO, DEPRECATED
        # Can be
        # - "host": "127.0.0.1"
        # - "host": "127.0.0.1,127.0.0.1"
        "host": "127.0.0.1",
        "port": 3306,
        "database": "mysql",
        "user": "your_account",
        "password": "your_password",
        "autocommit": True,
        "encoding": "utf8",
        # Pool
        "pool_max_size": 10,
    }

    # "unix" or "host" (not both), "host" has precedence
    # FOR : pysolmysql.Pool.mysql_pool.MysqlConnectionPool#_get_connection
    INTERNAL_SAMPLE_CONFIG_DICT = {
        "host": "127.0.0.1",
        "unix": "/var/run/mysqld/mysqld.sock",
        "port": 3306,
        "database": "mysql",
        "user": "your_account",
        "password": "your_password",
        "autocommit": True,
        "encoding": "utf8",
    }

    def __init__(self, conf_dict):
        """
        Init
        :param conf_dict: dict
        :type conf_dict: dict
        """

        Meters.aii("k.db_pool.mysql.call.__init")

        # Base
        super(MysqlConnectionPool, self).__init__(conf_dict)

        # Init us
        self.host_status = dict()

        # Check
        if "hosts" not in self.conf_dict and "host" not in self.conf_dict and "unix" not in self.conf_dict:
            raise Exception("No server specified (hosts, host, unix not found in conf_dict")

        # Init, "hosts" first
        if "hosts" in self.conf_dict:
            for host in self.conf_dict["hosts"]:
                self.host_status[host] = 0.0
        elif "host" in self.conf_dict:
            logger.warning("Using deprecated entry, prefer using 'hosts', got host=%s", self.conf_dict["host"])
            for host in self.conf_dict["host"].split(","):
                self.host_status[host] = 0.0
        elif "unix" in self.conf_dict:
            for host in self.conf_dict["unix"].split(","):
                self.host_status[host] = 0.0

    # ------------------------------------------------
    # HELPERS
    # ------------------------------------------------

    @classmethod
    def _get_connection(cls, conf_dict):
        """
        Get a connection
        :param conf_dict: dict
        :type conf_dict: dict
        :return: pymysql.connections.Connection
        :rtype: pymysql.connections.Connection
        """

        Meters.aii("k.db_pool.mysql.call._get_connection")

        # DOC :
        #     def __init__(self, host=None, user=None, password="",
        #          database=None, port=3306, unix_socket=None,
        #          charset='', sql_mode=None,
        #          read_default_file=None, conv=decoders, use_unicode=None,
        #          client_flag=0, cursorclass=Cursor, init_command=None,
        #          connect_timeout=None, ssl=None, read_default_group=None,
        #          compress=None, named_pipe=None, no_delay=None,
        #          autocommit=False, db=None, passwd=None, local_infile=False,
        #          max_allowed_packet=16*1024*1024, defer_connect=False,
        #          auth_plugin_map={}):

        logger.debug("mysql connect, server=%s:%s, unix=%s, user=%s, db=%s, enc=%s",
                     conf_dict.get("host"), conf_dict.get("port"), conf_dict.get("unix"),
                     conf_dict.get("user"),
                     conf_dict.get("database"),
                     conf_dict.get("encoding", "utf8")
                     )

        # Host
        h = conf_dict.get("host")

        # Unix detection (IF host startswith "/", we assume unix, even if unix not specified)
        if h and not h.startswith("/"):
            c = pymysql.connect(
                host=conf_dict["host"],
                port=int(conf_dict["port"]),

                db=conf_dict["database"],

                user=conf_dict["user"],
                password=conf_dict["password"],

                autocommit=conf_dict["autocommit"],

                charset=conf_dict.get("encoding", "utf8"),

                cursorclass=pymysql.cursors.DictCursor
            )
            return c

        # Unix
        h = conf_dict.get("unix")
        if not h:
            # Fallback host if unix format
            h = conf_dict.get("host")
            # Must start with "/"
            if not h.startswith("/"):
                h = None

        # Unix try
        if h:
            c = pymysql.connect(
                unix_socket=conf_dict["unix"],

                db=conf_dict["database"],

                user=conf_dict["user"],
                password=conf_dict["password"],

                autocommit=conf_dict["autocommit"],

                cursorclass=pymysql.cursors.DictCursor
            )
            return c

        # Nothing
        return None

    def _get_random_host(self):
        """
        Return a host in HOSTS_STATUS where the host is up
        :return str,bool (False)
        :rtype str,bool
        """
        now = time.time()
        hosts_up = [host for host, prison in self.host_status.items() if prison < now]
        try:
            host = random.choice(hosts_up)
            return host
        except IndexError:
            return False

    # ------------------------------------------------
    # OVERRIDES
    # ------------------------------------------------

    def _connection_create(self):
        """
        Create a connection, trying all available hosts if required (and disabling them if required)

        - conf_dict["hosts"] = ["host1", "host2"]

        where each host entry can be :
        - ip address (127.0.0.1)
        - host name (localhost)
        - unix socket name (/var/run/mysqld/mysqld.sock)

        :return: pymysql.connections.Connection
        :rtype: pymysql.connections.Connection
        """

        Meters.aii("k.db_pool.mysql.call._connection_create")

        # ------------------------
        # Try to get a connection
        # ------------------------
        out_conn = None
        while out_conn is None:
            # Pick a host
            host = self._get_random_host()

            # Check it
            if not host:
                Meters.aii("k.db_pool.mysql.hosts.all_down")
                raise Exception("No mysql host available, %s are down" % self.host_status.keys())

            # This host seems up => try open a connection
            try:
                # Build the dict for underlying _get_connection
                d_local = copy.deepcopy(self.conf_dict)
                if "hosts" in d_local:
                    del d_local["hosts"]
                d_local["host"] = host

                # Open it
                out_conn = self._get_connection(d_local)

                # Ping it (underlying base pool do NOT do it when opening connection)
                if not self._connection_ping(out_conn):
                    raise Exception("Connection ping failed")
            except Exception as e:
                # NOTE :
                # - We disable the host for ALL errors (even if DatabaseError can be raised when database do not exists but when the server is up...)

                # Deactivate host
                Meters.aii("k.db_pool.mysql.hosts.deactivate_one")
                logger.error("Host de-activate for 1 minute, host=%s, ex=%s", host, SolBase.extostr(e))
                self.host_status[host] = time.time() + 60.0
                # Kick connection
                self._connection_close(out_conn)

                # Reset
                out_conn = None

        # Over
        return out_conn

    def _connection_ping(self, conn):
        """
        Ping connection

        This send a ping, write/read toward mysql.

        :param conn: pymysql.connections.Connection
        :type conn: pymysql.connections.Connection
        :return bool
        :rtype bool
        """

        Meters.aii("k.db_pool.mysql.call._connection_ping")

        # noinspection PyBroadException
        try:
            conn.ping(reconnect=False)
        except Exception as e:
            Meters.aii("k.db_pool.mysql.ex_ping")
            logger.warning("Ping failed, obj=%s, ex=%s", conn, SolBase.extostr(e))
            return False
        else:
            return True

    def _connection_close(self, conn):
        """
        Close a connection
        Must not raise anything.
        :param conn: pymysql.connections.Connection
        :type conn: pymysql.connections.Connection
        """

        Meters.aii("k.db_pool.mysql.call._connection_close")

        # noinspection PyBroadException
        try:
            if conn:
                conn.close()
        except Exception as e:
            # Dont care of exception in case of closing
            Meters.aii("k.db_pool.mysql.ex_close")
            logger.debug("Close exception (non fatal), ex=%s", SolBase.extostr(e))



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""

import logging
from gevent import queue
from gevent.threading import Lock
from pysolmeters.Meters import Meters

logger = logging.getLogger(__name__)


class DatabaseConnectionPool(object):
    """
    Connection pool.
    """

    def __init__(self, conf_dict):
        """
        Init
        :param conf_dict: dict
        :type conf_dict: dict
        """

        # Lock for acquire/release
        self.pool_lock = Lock()

        # Store
        self.conf_dict = conf_dict

        # Max size
        self.max_size = self.conf_dict.get("pool_max_size", 10)

        # Alloc
        self.pool = queue.Queue(maxsize=self.max_size)

        # Init
        self.size = 0

    def connection_acquire(self):
        """
        Get a connection
        :return: object
        :rtype object
        """

        with self.pool_lock:

            Meters.aii("k.db_pool.base.call.connection_acquire")

            if self.pool.qsize() > 0:
                # ------------------------------
                # GET CONNECTION FROM POOL
                # ------------------------------
                conn = self.pool.get()

                # Ping it
                if not self._connection_ping(conn):
                    # Failed => close it
                    self._connection_close(conn)

                    # Re-create a new one (we just closed a connection)
                    conn = self._connection_create()

                # Send it back
                return conn
            elif self.size >= self.max_size:
                # ------------------------------
                # POOL MAXED => ERROR
                # ------------------------------
                Meters.aii("k.db_pool.base.pool_maxed")
                raise Exception("Pool maxed, size=%s, max_size=%s" % (self.size, self.max_size))
            else:
                # ------------------------------
                # POOL NOT MAXED, NO CONNECTION IN POOL => NEW CONNECTION
                # ------------------------------
                try:
                    conn = self._connection_create()
                    self.size += 1
                    Meters.aii("k.db_pool.base.cur_size", increment_value=1)
                    Meters.ai("k.db_pool.base.max_size").set(max(Meters.aig("k.db_pool.base.max_size"), Meters.aig("k.db_pool.base.cur_size")))
                except Exception:
                    raise
                return conn

    def connection_release(self, conn):
        """
        Put a connection back in the pool
        :param conn: object
        :type conn: object
        """

        with self.pool_lock:

            Meters.aii("k.db_pool.base.call.connection_release")

            # If none, decrement + return
            if conn is None:
                return

            # Put it back
            try:
                self.pool.put(conn)
            except queue.Full:
                # If full, close it
                self._connection_close(conn)

    def close_all(self):
        """
        Close all connections
        """
        n = 0
        while not self.pool.empty():
            conn = self.pool.get_nowait()
            self._connection_close(conn)
            n += 1

        Meters.aii("k.db_pool.base.cur_size", increment_value=-n)
        Meters.ai("k.db_pool.base.max_size").set(max(Meters.aig("k.db_pool.base.max_size"), Meters.aig("k.db_pool.base.cur_size")))
        self.size = 0

    # ------------------------------------------------
    # OVERRIDES
    # ------------------------------------------------

    def _connection_create(self, *args, **kwargs):
        """
        Create connection
        :param args: object
        :type args: object
        :param kwargs: object
        :type kwargs: object
        :return object
        :rtype object
        """
        raise NotImplementedError("create_connection")

    def _connection_ping(self, conn):
        """
        Ping connection
        :param conn: object
        :type conn: object
        :return bool
        :rtype bool
        """

        raise NotImplementedError("ping_connection")

    def _connection_close(self, conn):
        """
        Close a connection.
        Must not raise anything.
        :param conn: object
        :type conn: object
        """

        raise NotImplementedError("close_connection")



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""

# This is forked and updated from : https://github.com/laurentL/django-mysql-geventpool-27

# Which was forked : https://github.com/shunsukeaihara/django-mysql-geventpool

# NOTE :
# we do not close connection or timeout them, we assume underlying backend (mariadb) will close inactive connections on its end

# TODO : handle initial connection opening (for warm-up the pool when allocating)
# TODO : handle connection_acquire with timeout (in case pool is full - currently we raise immediatly)
# TODO : handle better locking in connection_acquire (global lock currently)



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""
from distutils.core import setup

import re
from setuptools import find_packages


def requirement_read(req_file):
    """
    Doc
    :param req_file: Doc
    :return: Doc
    """
    local_req_list = list()
    local_dep_list = list()
    for row_buffer in open(req_file).readlines():
        # Skip empty
        if len(row_buffer.strip()) == 0:
            continue
        # Skip "- ..."
        elif re.match("^-", row_buffer):
            continue
        # Skip "# ..."
        elif re.match("^#", row_buffer):
            continue
        # Git stuff (direct)
        # From : https://github.com/pypa/pip/issues/3610#issuecomment-356687173
        elif re.match("^git", row_buffer):
            pkg = row_buffer.split('#')[-1]
            local_dep_list.append(row_buffer.strip() + '-9876543210')
            local_req_list.append(pkg.replace('egg=', '').rstrip())
        else:
            # Ok
            local_req_list.append(row_buffer)

    print("local_req_list={0}".format(local_req_list))
    print("local_dep_list={0}".format(local_dep_list))
    return local_req_list, local_dep_list


# ===========================
# SETUP
# ===========================

p_name = "pysolmysql"
p_author = "Laurent Champagnac"
p_email = "champagnac.laurent@gmail.com"
p_url = "https://knock.center"
p_version = "1.1.5"

# Load
req_list, dep_list = requirement_read("requirements.txt")
test_req_list, _ = requirement_read("requirements_test.txt")

setup(

    # Project details
    name=p_name,
    author=p_author,
    author_email=p_email,
    url=p_url,
    description="Simple MYSQL client Apis",

    # Version, format : Major.Minor.Revision
    version=p_version,

    # Packages
    packages=find_packages(exclude=["*_test*", "_*"]),
    include_package_data=True,

    # License & read me
    license="GPLv3",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",

    # Data files
    data_files=[
        ("", ["requirements_test.txt", "requirements.txt", "README.md", "LICENSE.md"]),
    ],

    # Classifiers
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries",
        "Natural Language :: English",
    ],

    # Dependencies
    install_requires=req_list,

    # Direct deps
    dependency_links=dep_list,

    # Dependencies : test
    tests_require=test_req_list,

    # Zip
    zip_safe=False,
)
